from abc import ABC
from typing import Callable, Optional

from orinoco import config
from orinoco.action import (
    Action,
    record_action,
    verbose_action_exception,
    async_record_action,
    async_verbose_action_exception,
)
from orinoco.entities import ActionData, Signature
from orinoco.exceptions import ActionNotReturnedActionData, ActionNotProperlyInherited
from orinoco.types import ActionDataT


class Transformation(Action, ABC):
    """
    Base class for actions which modify data
    """

    @record_action
    @verbose_action_exception
    def run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        return self._check_transformation_output(self.transform(action_data))

    @async_record_action
    @async_verbose_action_exception
    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        return self._check_transformation_output(await self.async_transform(action_data))

    def transform(self, action_data: ActionDataT) -> ActionDataT:
        raise ActionNotProperlyInherited(
            "`transform` method needs to implemented for sync execution or `async_transform` for async execution"
        )

    async def async_transform(self, action_data: ActionDataT) -> ActionDataT:
        return self.transform(action_data)

    @classmethod
    def _check_transformation_output(cls, action_data: ActionDataT) -> ActionDataT:
        if config.CHAINING_TYPE_CHECK_STRICT_MODE_ENABLED and not isinstance(action_data, ActionData):
            raise ActionNotReturnedActionData(
                "Action {} not returned ActionData, but {}".format(action_data, cls.__name__)
            )
        return action_data


class GenericTransformation(Transformation):
    """
    Utility action which allows modification of `~orinoco.entities.ActionData` "on the fly"
    (usually by lambda functions)
    """

    def __init__(
        self,
        method: Callable[[ActionDataT], ActionDataT],
        description: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        :param method: Method which takes `ActionData`, modify them and return new copy
        :param description:
        """
        super().__init__(description=description, name=name)
        self.method = method

    def transform(self, action_data: ActionDataT) -> ActionDataT:
        action_data_result = self.method(action_data)

        if config.CHAINING_TYPE_CHECK_STRICT_MODE_ENABLED and not isinstance(action_data_result, ActionData):
            raise ActionNotReturnedActionData(
                "Action {} not returned ActionData, but {}".format(action_data_result, self.__class__.__name__)
            )

        return action_data_result


class RenameActionField(Transformation):
    """
    Change name of the registered entity
    """

    def __init__(self, key: str, new_key: str):
        super().__init__(name="{}[{} -> {}]".format(self.__class__.__name__, key, new_key))
        self.key = key
        self.new_key = new_key

    def transform(self, action_data: ActionDataT) -> ActionDataT:
        new_data = []
        for signature, value in action_data.data:
            if signature.key == self.key:
                new_data.append((signature.evolve_self(key=self.new_key), value))
            else:
                new_data.append((signature, value))
        return action_data.evolve_self(data=new_data)


class WithoutFields(Transformation):
    """
    Remove signatures for the given keys
    """

    def __init__(self, *fields: str):
        super().__init__(name="{}[{}]".format(self.__class__.__name__, ", ".join(fields)))
        self.fields = fields

    def transform(self, action_data: ActionDataT) -> ActionDataT:
        return action_data.remove_many(
            searched_signatures=[Signature(key=field) for field in self.fields],
            exact_match=False,
            ignore_non_existent=False,
        )



from abc import ABC, abstractmethod
from typing import Iterable, Any, Callable, Optional, TypeVar, Generic, AsyncIterable

from orinoco.action import (
    Action,
    ActionSet,
    record_action,
    verbose_action_exception,
    async_record_action,
    async_verbose_action_exception,
)

from orinoco.exceptions import ActionNotProperlyConfigured, RunnableOnlyInAsyncContext
from orinoco.types import ActionDataT

LoopT = TypeVar("LoopT", bound="BaseLoop")


class BaseLoop(Generic[LoopT], Action, ABC):
    def __init__(self, iterating_key: str, action: Optional[Action] = None):
        """
        :param iterating_key: Key which will be propagated into the :class:`~orinoco.entities.ActionData` with the
        new value
        """
        super().__init__(description="Loop over '{}'".format(iterating_key))
        self.iterating_key = iterating_key

        self.action = action

    def do(self: LoopT, *actions: Action) -> LoopT:
        """
        :param actions: Actions which are executed in the each iteration with updated
         values in the  :class:`~orinoco.entities.ActionData`
        :return:
        """
        self.action = ActionSet(actions)
        return self


class LoopEvents(BaseLoop):
    """
    Loop implementation for actions which runs just side-effects. Values obtained in the loop are
    not propagated further in the actions chain
    """

    @record_action
    @verbose_action_exception
    def run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        if not self.action:
            raise ActionNotProperlyConfigured(
                "Action for the loops is not set. Call `do` method first or provide it through the constructor"
            )
        for loop_data in self._get_data_generator(action_data):
            self.action.run(loop_data)
        return action_data

    def _get_data_generator(self, action_data: ActionDataT) -> Iterable[ActionDataT]:
        for value in self._get_generator(action_data):
            yield action_data.evolve(**{self.iterating_key: value})

    @abstractmethod
    def _get_generator(self, action_data: ActionDataT) -> Iterable[Any]:
        """
        :param action_data: Data container passed by actions
        :return: Iterable which provides the value which is set to the :class:`~orinoco.entities.ActionData`
        in each of the iterations
        """
        pass


class ForSideEffects(LoopEvents):
    """
    Implementation of `Loop` which uses a method for generating loop iterable

    Example:

    .. code-block:: python

        ForSideEffects("event", lambda action_data: action_data.get_or_default("event_log", [])).do(GetNotification())

    Then `GetNotification` has access to `event` via action data in each of the iterations
    which comes from the iterable "event_log"
    """

    def __init__(self, iterating_key: str, method: Callable[[ActionDataT], Iterable[Any]]):
        """
        :param iterating_key: Key which will be propagated into the `ActionData` with the new value
        :param method: Method which returns the iterable to iterate over
        """
        super().__init__(iterating_key=iterating_key)
        self.method = method

    def _get_generator(self, action_data: ActionDataT) -> Iterable[ActionDataT]:
        return self.method(action_data)


class For(BaseLoop):
    """
    Implementation of `Loop` which uses a method for generating loop iterable

    Example:

    .. code-block:: python

        For("x", lambda ad: ad.get("values"), aggregated_field="doubled", aggregated_field_new_name="doubled_list")
        .do(DoubleValue())
        .run_with_data(values=[10, 40, 60])
        .get("doubled_list")
    """

    def __init__(
        self,
        iterating_key: str,
        method: Callable[[ActionDataT], Iterable[Any]],
        aggregated_field: Optional[str] = None,
        aggregated_field_new_name: Optional[str] = None,
        skip_none_for_aggregated_field: bool = False,
    ):
        """
        :param iterating_key: Key which will be propagated into the `ActionData` with the new value
        :param method: Method which returns the iterable to iterate over
        :param aggregated_field: Name of the field which will be extracted from the `ActionData` and aggregated
        (appended to the list)
        :param aggregated_field_new_name: Name of the field which will be used for the aggregated field
        :param skip_none_for_aggregated_field: If `True` then `None` values won't be added to the aggregated field
        """
        super().__init__(iterating_key=iterating_key)
        self.method = method
        self.aggregated_field = aggregated_field
        self.aggregated_field_new_name = aggregated_field_new_name
        self.skip_none_for_aggregated_field = skip_none_for_aggregated_field

    @record_action
    @verbose_action_exception
    def run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        if not self.action:
            raise ActionNotProperlyConfigured(
                "Action for the loops is not set. Call `do` method first or provide it through the constructor"
            )

        aggregated_values = []
        for iteration_value in self.method(action_data):
            loop_action_data = self.action.run(action_data.evolve(**{self.iterating_key: iteration_value}))
            if self.aggregated_field:
                value_to_aggregate = action_data.get(self.aggregated_field)
                if value_to_aggregate is not None or (not self.skip_none_for_aggregated_field):
                    aggregated_values.append(value_to_aggregate)

        if self.aggregated_field:
            return loop_action_data.evolve(
                **{self.aggregated_field_new_name or self.aggregated_field: aggregated_values}
            )
        return action_data


class AsyncLoopEvents(LoopEvents):
    """
    Loop implementation for actions which runs just side-effects/conditions. Values obtained in the loop are
    not propagated further in the actions chain
    """

    @async_record_action
    @async_verbose_action_exception
    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        if not self.action:
            raise ActionNotProperlyConfigured(
                "Action for the loops is not set. Call `do` method first or provide it through the constructor"
            )

        async for loop_data in self._get_async_data_generator(action_data):
            await self.action.async_run(loop_data)
        return action_data

    async def _get_async_data_generator(self, action_data: ActionDataT) -> AsyncIterable[ActionDataT]:
        async for value in self._get_generator(action_data):
            yield action_data.evolve(**{self.iterating_key: value})

    @abstractmethod
    def _get_generator(self, action_data: ActionDataT) -> AsyncIterable[Any]:
        """
        :param action_data: Data container passed by actions
        :return: Iterable which provides the value which is set to the :class:`~orinoco.entities.ActionData`
        in each of the iterations
        """
        pass


class AsyncForSideEffects(AsyncLoopEvents):
    """
    Implementation of `Async Loop` which uses a method for generating loop async iterable

    Example:

    .. code-block:: python

        For("event", lambda action_data: action_data.get_or_default("event_log", [])).do(GetNotification())

    Then `GetNotification` has access to `event` via action data in each of the iterations
    which comes from the async iterable "event_log"
    """

    def __init__(self, iterating_key: str, method: Callable[[ActionDataT], AsyncIterable[Any]]):
        """
        :param iterating_key: Key which will be propagated into the `ActionData` with the new value
        :param method: Method which returns the iterable to iterate over
        """
        super().__init__(iterating_key=iterating_key)
        self.method = method

    def _get_generator(self, action_data: ActionDataT) -> AsyncIterable[ActionDataT]:
        return self.method(action_data)


class AsyncFor(BaseLoop):
    def run(self, action_data: ActionDataT) -> ActionDataT:
        raise RunnableOnlyInAsyncContext()

    def __init__(
        self,
        iterating_key: str,
        method: Callable[[ActionDataT], AsyncIterable[Any]],
        aggregated_field: Optional[str] = None,
        aggregated_field_new_name: Optional[str] = None,
        skip_none_for_aggregated_field: bool = False,
    ):
        """
        :param iterating_key: Key which will be propagated into the `ActionData` with the new value
        :param method: Method which returns the iterable to iterate over
        :param aggregated_field: Name of the field which will be extracted from the `ActionData` and aggregated
        (appended to the list)
        :param aggregated_field_new_name: Name of the field which will be used for the aggregated field
        :param skip_none_for_aggregated_field: If `True` then `None` values won't be added to the aggregated field
        """
        super().__init__(iterating_key=iterating_key)
        self.method = method
        self.aggregated_field = aggregated_field
        self.aggregated_field_new_name = aggregated_field_new_name
        self.skip_none_for_aggregated_field = skip_none_for_aggregated_field

    @async_record_action
    @async_verbose_action_exception
    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        if not self.action:
            raise ActionNotProperlyConfigured(
                "Action for the loops is not set. Call `do` method first or provide it through the constructor"
            )

        aggregated_values = []
        async for iteration_value in self.method(action_data):
            action_data = await self.action.async_run(action_data.evolve(**{self.iterating_key: iteration_value}))
            if self.aggregated_field:

                value_to_aggregate = action_data.get(self.aggregated_field)
                if value_to_aggregate is not None or (not self.skip_none_for_aggregated_field):
                    aggregated_values.append(value_to_aggregate)

        if self.aggregated_field:
            return action_data.evolve(**{self.aggregated_field_new_name or self.aggregated_field: aggregated_values})
        return action_data



from __future__ import annotations

from abc import ABC
from typing import Optional, Dict, Any, Type, Callable, Generic, Awaitable, TypeVar

from orinoco import config
from orinoco.action import (
    Action,
    record_action,
    verbose_action_exception,
    async_record_action,
    async_verbose_action_exception,
)
from orinoco.condition import Condition
from orinoco.entities import ActionConfig, Signature
from orinoco.exceptions import ActionNotProperlyConfigured
from orinoco.helpers import extract_type
from orinoco.retry import WaitUntilTrue, WaitUntilEqualsTo, WaitUntilContains, WaitUntilNotFail
from orinoco.types import T, ActionDataT

TypedBaseT = TypeVar("TypedBaseT", bound="TypedBase")


class TypedBase(Generic[T], Action, ABC):
    CONFIG: Optional[ActionConfig[T]] = None
    config: ActionConfig[T]

    __call__: Callable[..., Any]

    def __init__(
        self, description: Optional[str] = None, config: Optional[ActionConfig[T]] = None, name: Optional[str] = None
    ):
        Action.__init__(self, description=description, name=name)

        self.config = config or self.CONFIG or self._get_implicit_config()

    def get_input_params(self, action_data: ActionDataT) -> Dict[str, Any]:
        return {key: action_data.find_one(signature) for key, signature in (self.config.INPUT or {}).items()}

    def output_as(self: TypedBaseT, key: str, type_: Type | None = None) -> TypedBaseT:
        """
        Mutable change of the output signature of the action.
        """
        self.config = self.config.evolve_self(OUTPUT=Signature(key=key, type_=type_))
        return self

    @classmethod
    def _get_implicit_config(cls) -> ActionConfig[T]:
        annotations = cls.__call__.__annotations__
        return_type, return_name, tags = extract_type(annotations["return"])

        if config.IMPLICIT_TYPE_STRICT_MODE_ENABLED and return_name is None:
            raise ActionNotProperlyConfigured(
                "Action {} has to be configured explicitly or return type has to be annotated via "
                "`Annotated[<type>, <name>]`. The error was raised, because `IMPLICIT_TYPE_STRICT_MODE_ENABLED` "
                "is enabled.".format(cls)
            )

        return ActionConfig(
            OUTPUT=Signature(type_=return_type, key=return_name, tags=tags) if return_type else None,
            INPUT={key: Signature(key=key) for key, type_ in annotations.items() if key != "return"},
        )

    def retry_until_not_fails(
        self, exception_cls_to_catch=BaseException, max_retries: int = 10, retry_delay: float = 10
    ) -> WaitUntilNotFail:
        return WaitUntilNotFail(
            action=self, exception_cls_to_catch=exception_cls_to_catch, max_retries=max_retries, retry_delay=retry_delay
        )


class TypedActionBase(Generic[T], TypedBase[T], ABC):
    def retry_until_equals(self, value: Any, max_retries: int = 10, retry_delay: float = 10) -> WaitUntilEqualsTo:
        if self.config.OUTPUT is None or self.config.OUTPUT.key is None:
            raise ActionNotProperlyConfigured("Retry until condition has to be annotated with the name of the output")

        return WaitUntilEqualsTo(
            self, key=self.config.OUTPUT.key, value=value, max_retries=max_retries, retry_delay=retry_delay
        )

    def retry_until_contains(self, value: Any, max_retries: int = 10, retry_delay: float = 10) -> WaitUntilContains:
        if self.config.OUTPUT is None or self.config.OUTPUT.key is None:
            raise ActionNotProperlyConfigured("Retry until condition has to be annotated with the name of the output")
        return WaitUntilContains(
            self, key=self.config.OUTPUT.key, value=value, max_retries=max_retries, retry_delay=retry_delay
        )


class TypedAction(Generic[T], TypedActionBase[T], ABC):
    """
    Enhanced `Action` which use `ActionConfig` as configuration of the input and the output.

    Business logic is defined in the call method with "normal" parameters as the input, this means that no raw
    `ActionData` is required. The propagation of the data from the `ActionData` to the method is done automatically
    based on the `ActionConfig` which can be passed directly to the constructor (see `config`) or as the class variable
    (see `CONFIG`) or implicitly from annotations of the `__call__` method.

    The result of the method is propagated to the `ActionData` with the corresponding signature. Note that implicit
    config uses only annotated return type for the signature. For more control define the `ActionConfig` manually.
    """

    __call__: Callable[..., T]

    @record_action
    @verbose_action_exception
    def run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        result = self(**self.get_input_params(action_data))

        if self.config.OUTPUT:
            return action_data.register(signature=self.config.OUTPUT, entity=result)
        return action_data


class AsyncTypedAction(Generic[T], TypedActionBase[T], ABC):
    """
    Async version of :class:`TypedAction`
    """

    SYNC_ACTION: Optional[Type[TypedAction[T]]] = None

    __call__: Callable[..., Awaitable[T]]

    def __init__(
        self,
        description: Optional[str] = None,
        config: Optional[ActionConfig[T]] = None,
        name: Optional[str] = None,
        sync_action_kwargs: Optional[Dict[str, Any]] = None,
    ):
        super(AsyncTypedAction, self).__init__(description=description, config=config, name=name)

        self.sync_action = self.SYNC_ACTION(**(sync_action_kwargs or {})) if self.SYNC_ACTION else None

    @async_record_action
    @async_verbose_action_exception
    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        result = await self(**self.get_input_params(action_data))

        if self.config.OUTPUT:
            return action_data.register(signature=self.config.OUTPUT, entity=result)
        return action_data

    @record_action
    @verbose_action_exception
    def run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        if not self.sync_action:
            raise ActionNotProperlyConfigured(
                "SYNC_ACTION action need to be provided in order to run this async action synchronously"
            )
        return self.sync_action.run(action_data)


class TypedCondition(Condition, TypedBase[bool], ABC):
    """
    Condition base of :class:`TypedAction` type
    """

    __call__: Callable[..., bool]

    def __init__(
        self,
        fail_message: Optional[str] = None,
        is_inverted: bool = False,
        error_cls: Optional[Type[Exception]] = None,
        config: Optional[ActionConfig[bool]] = None,
        name: Optional[str] = None,
    ):
        TypedBase.__init__(self, config=config)
        Condition.__init__(self, fail_message=fail_message, is_inverted=is_inverted, error_cls=error_cls, name=name)

    def retry_until(self, max_retries: int = 10, retry_delay: float = 10) -> WaitUntilTrue:
        return WaitUntilTrue(self, max_retries=max_retries, retry_delay=retry_delay)

    def _is_valid(self, action_data: ActionDataT) -> bool:
        return self(**self.get_input_params(action_data))



from pydantic import BaseSettings


class Config(BaseSettings):
    IMPLICIT_TYPE_STRICT_MODE_ENABLED: bool = False
    CHAINING_TYPE_CHECK_STRICT_MODE_ENABLED: bool = False
    VERBOSE_ERRORS = False

    class Config:
        env_prefix = "ORINOCO_"



from abc import ABC
from typing import Any, Callable, Union, Optional

from orinoco.action import (
    Action,
    record_action,
    verbose_action_exception,
    ActionSet,
    async_record_action,
    async_verbose_action_exception,
)
from orinoco.entities import ActionData
from orinoco.exceptions import ActionNotProperlyInherited
from orinoco.helpers import raise_not_provided_field
from orinoco.types import ActionDataT


class DataSource(Action, ABC):
    """
    Base action which provides new data
    """

    PROVIDES: Optional[str] = None

    def __init__(
        self,
        provides: Optional[str] = None,
        dont_get_if_is_in: bool = False,
        description: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        :param provides: Key in :class:`~orinoco.entities.ActionData` under the data it will be stored.
        If it's not provided :attrs:`orinoco.data_source.DataSource.PROVIDES` is used as default value.
        :param dont_get_if_is_in: If it's True and a value with ``provides`` key is already
        in `:class:`~orinoco.entities.ActionData` the :func:`~orinoco.data_source.DataSource.get_data` is not executed.
        :param description: Description of the action
        :param name: Name of the action
        """
        super().__init__(name=name, description=description)
        self.provides = provides or self.PROVIDES or raise_not_provided_field("provides")
        self.dont_get_if_is_in = dont_get_if_is_in

    @record_action
    @verbose_action_exception
    def run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        if self.dont_get_if_is_in and action_data.is_in(self.provides):
            return action_data
        return action_data.evolve(**{self.provides: self.get_data(action_data)})

    @async_record_action
    @async_verbose_action_exception
    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        if self.dont_get_if_is_in and action_data.is_in(self.provides):
            return action_data
        return action_data.evolve(**{self.provides: await self.async_get_data(action_data)})

    def get_data(self, action_data: ActionDataT) -> Any:
        """
        :param action_data:
        :return: Value which is appended to `ActionData`
        """
        raise ActionNotProperlyInherited(
            "`get_data` method needs to implemented for sync execution or `async_get_data` for async execution"
        )

    async def async_get_data(self, action_data: ActionDataT) -> Any:
        """
        :param action_data: Data container passed by actions
        :return: Value which is appended to `~orinoco.entities.ActionData`
        """
        return self.get_data(action_data)

    def get_source_data(self, **kwargs: Any) -> Any:
        """
        Shortcut to call the :class:`~orinoco.data_source.DataSource` instance directly and get the result

        :param kwargs: Data which are propagated into `~orinoco.entities.ActionData`
        :return: Requested data value
        """
        return self.get_data(ActionData.create(**kwargs))


class GenericDataSource(DataSource):
    """
    Utility data source action which allows inserting values into the `~orinoco.entities.ActionData` "on the fly"
    (usually by lambda functions)
    """

    def __init__(
        self, method: Callable[[ActionDataT], Any], provides: Optional[str] = None, name: Optional[str] = None
    ):
        """
        :param method: Method which use :class:`~orinoco.entities.ActionData` to get the data
        :param provides:
        """
        super().__init__(provides=provides, name=name)
        self.method = method

    def get_data(self, action_data: ActionDataT) -> Any:
        return self.method(action_data)


class AddActionValue(DataSource):
    """
    Utility class which adds a value into the :class:`~orinoco.entities.ActionData`
    """

    def __init__(self, key: str, value: Union[Any, Callable[[], Any]]):
        """
        :param key: Key of the value in the :class:`~orinoco.entities.ActionData`
        :param value: Value to set
        """
        super().__init__(provides=key, name="{}[{}]".format(self.__class__.__name__, key))
        self.value = value

    def get_data(self, action_data: ActionDataT) -> Any:
        if callable(self.value):
            return self.value()

        return self.value


class AddActionValues(ActionSet):
    def __init__(self, **action_values: Any):
        super().__init__([AddActionValue(k, v) for k, v in action_values.items()])


class AddVirtualKeyShortcut(GenericDataSource):
    def __init__(self, key: str, source_key: str):
        """
        :param key: New key in the :class:`~orinoco.entities.ActionData`
        :param source_key: Key for getting the value from :class:`~orinoco.entities.ActionData`
        """
        super().__init__(
            provides=key,
            method=lambda ad: ad.get(source_key),
            name="{}[{} -> {}]".format(self.__class__.__name__, source_key, key),
        )



import abc
from functools import partial
from typing import Dict, Any, Optional, Set, Type, List, Tuple, Sequence, ClassVar, Generic, Awaitable, Union

from pydantic import Field

from orinoco.exceptions import SearchError, NothingFound, FoundMoreThanOne, AlreadyRegistered
from orinoco.helpers import initialize
from orinoco.observers import ExecutionTimeObserver, ActionsLog
from orinoco.types import (
    SignatureT,
    BaseModelT,
    T,
    ActionConfigT,
    ActionDataT,
    ActionT,
    ObserverT,
    ObserverVar,
    ImmutableEvolvableModelT,
)


class ImmutableEvolvableModel(ImmutableEvolvableModelT, abc.ABC):
    def evolve_self(self: BaseModelT, **kwargs: Any) -> BaseModelT:
        return self.copy(update=kwargs)


class Signature(Generic[T], ImmutableEvolvableModel, SignatureT[T]):
    type_: Optional[Type[T]] = None
    tags: Set[str] = Field(default_factory=set)
    key: Optional[str] = None

    def match(self, other_signature: "SignatureT[T]") -> bool:
        """
        Match with another signature. Note that matching with for ``tags`` set is done as "is subset"
        """
        if (
            other_signature.type_ is not None
            and self.type_ is not None
            and not issubclass(self.type_, other_signature.type_)
        ):
            return False

        if len(other_signature.tags) > 0 and not other_signature.tags.issubset(self.tags):
            return False

        if other_signature.key is not None and other_signature.key != self.key:
            return False
        return True

    def __class_getitem__(cls: T, _: Any) -> T:
        # Fix for pydantic to support generic types (expressions like `SignatureT[bool]`)
        return cls


class ActionConfig(Generic[T], ImmutableEvolvableModel, ActionConfigT[T]):
    INPUT: Optional[Dict[str, SignatureT[Any]]] = None
    OUTPUT: Optional[SignatureT[T]] = None

    @classmethod
    def create_strict(
        cls,
        input_: Optional[Dict[str, Type[Any]]] = None,
        output_type: Optional[Type[T]] = None,
        output_name: Optional[str] = None,
    ) -> "ActionConfig[T]":
        return cls(
            INPUT=({name: Signature(type_=type_, key=name) for name, type_ in input_.items()}) if input_ else None,
            OUTPUT=Signature(type_=output_type, key=output_name),
        )

    def __class_getitem__(cls: T, _: Any) -> T:
        # Fix for pydantic to support generic types (expressions like `SignatureT[bool]`)
        return cls


class ActionData(ImmutableEvolvableModel, ActionDataT):
    """
    Fundamental data structure which is both input and output of actions.

    Virtual keys: To save a memory some data are not stored directly, but just a method how to get them from other
    data stored in `ActionData`.

    `ActionData.data`, `ActionData.virtual_keys` and `ActionData` are immutable data structures, so usually a copy
    is made after every action it pass thru with the new data.

    History of all actions done on the dataset are recorded in `ActionData.actions_done`.
    """

    NOT_FOUND: ClassVar = object()
    DEFAULT_OBSERVERS_CLASSES: ClassVar = (ExecutionTimeObserver, ActionsLog)

    data: Tuple[Tuple[SignatureT[Any], Any], ...] = Field(default_factory=tuple)
    futures: List[Awaitable[Any]] = Field(default_factory=list)
    observers: List[ObserverT] = Field(default_factory=partial(initialize, DEFAULT_OBSERVERS_CLASSES))

    skip_processing: bool = False

    # Shortcuts
    def __getitem__(self, searched_signature: SignatureT[T]) -> T:
        return self.get_by_signature(searched_signature)

    def get(self, key: str, default: Any = NOT_FOUND) -> Any:
        """
        Get item by ``key``
        """
        return self.get_by_key(key, default=default)

    # Getters
    def get_by_signature(self, searched_signature: SignatureT[T]) -> T:
        """
        Get data with the given signature -- all non-default fields of the signature will be used for matching
        """
        return self.get_with_signature(searched_signature=searched_signature)[1]

    def get_with_signature(self, searched_signature: SignatureT[T]) -> Tuple[SignatureT[T], T]:
        """
        Get data + signature with the given signature -- all non-default fields of the signature will be used
        for matching
        """
        try:
            return self._ensure_one(
                [(signature, entity) for signature, entity in self.data if signature == searched_signature]
            )
        except SearchError as err:
            raise SearchError(
                "Failed to find {}\nPresent signatures: {}".format(searched_signature, self.signatures)
            ) from err

    def get_or_default(self, key: str, default: Optional[Any] = None) -> Any:
        """
        Get data by key or return a default value
        """
        try:
            return self.get_by_signature(Signature(key=key))
        except (NothingFound, SearchError):
            return default

    def find_or_default(self, key: str, default: Optional[Any] = None) -> Any:
        """
        Find a single value with matching or default
        """
        try:
            return self.find_one(Signature(key=key))
        except NothingFound:
            return default

    def get_by_key(self, key: str, default: Any = NOT_FOUND) -> Any:
        """
        Get item by ``key``
        """
        try:
            return self.find_one(Signature(key=key))
        except NothingFound:
            if "." in key:
                name_parts = key.split(".")
                result = self._get_from_nested(
                    key=".".join(name_parts[1:]),
                    data=self.find_one(Signature(key=name_parts[0])),
                    default=self.NOT_FOUND,
                )
                if result != self.NOT_FOUND:
                    return result

            if default != self.NOT_FOUND:
                return default
            raise

    def get_by_type(self, type_: Type[T]) -> T:
        """
        Get a value by matching type.

        :raises
         - `NothingFound`: When no data match the signature
         - `FoundMoreThanOne`: When multiple data match the signature
        """
        return self.find_one(Signature(type_=type_))

    def get_by_tags(self, *tags: str) -> Any:
        """
        Get a value by matching tags.

        :raises
         - `NothingFound`: When no data match the signature
         - `FoundMoreThanOne`: When multiple data match the signature
        """
        return self.find_one(Signature(tags=set(tags)))

    # Searching
    def find(self, searched_signature: SignatureT[T]) -> List[T]:
        """
        Find values which match the signature
        """
        return list(map(lambda signature_value: signature_value[1], self.find_with_signature(searched_signature)))

    def find_one(self, searched_signature: SignatureT[T]) -> T:
        """
        Find single value which match the signature

        :raises
         - `NothingFound`: When no data match the signature
         - `FoundMoreThanOne`: When multiple data match the signature
        """
        try:
            return self._ensure_one(self.find(searched_signature=searched_signature))
        except SearchError as err:
            raise err.__class__(
                "Failed to find {}\nPresent signatures: {}".format(searched_signature, self.signatures)
            ) from err

    def find_with_signature(self, searched_signature: SignatureT[T]) -> List[Tuple[SignatureT[T], T]]:
        """
        Find values with signatures which match the signature
        """
        return [(signature, entity) for signature, entity in self.data if signature.match(searched_signature)]

    # --- Evolution
    def evolve(self, **data: Any) -> "ActionData":
        """
        Add values to the data with signatures created from keys (parameter names)

        Already registered data with signatures with keys matching the new will be removed and replaced by the new ones
        """
        # TODO: Can be optimized
        existing = {}
        for key in data.keys():
            try:
                signature: SignatureT[Any] = self._ensure_one(self.find_with_signature(Signature(key=key)))[0]
                existing[key] = signature

            except SearchError:
                pass

        return self.remove_many(searched_signatures=list(existing.values()), ignore_non_existent=True).register_many(
            [(existing.get(key, Signature(key=key, type_=type(value))), value) for key, value in data.items()]
        )

    def register(self, signature: SignatureT[T], entity: T, check_if_exists: bool = True) -> "ActionData":
        """
        Add new value with the given signature
        :param signature: Signature of the entity
        :param entity: Entity which will be added
        :param check_if_exists: Controls whether the method raises an exception if a value with
        the ``signature`` exists
        :return: Copy of itself
        """
        if check_if_exists:
            try:
                self.get_by_signature(searched_signature=signature)
                raise AlreadyRegistered("Entity with signature {} is already registered".format(signature))
            except SearchError:
                pass

        # Remove data with an existing key
        instance = self
        if signature.key:
            instance = self.remove(
                searched_signature=Signature(key=signature.key), ignore_non_existent=True, exact_match=False
            )
        return instance.evolve_self(data=instance.data + ((signature, entity),))

    def register_many(self, data: Sequence[Tuple[SignatureT[T], T]], check_if_exists: bool = True) -> "ActionData":
        """
        Add new values with the given signatures
        :param data: Items with signatures to be added
        :param check_if_exists: Controls whether the method raises an exception if a value with
        the ``signature`` exists

        :return: Copy of itself
        """
        instance = self
        for signature, entity in data:
            instance = self.register(signature=signature, entity=entity, check_if_exists=check_if_exists)
        return instance

    def remove(
        self, searched_signature: SignatureT[Any], ignore_non_existent: bool = False, exact_match: bool = True
    ) -> "ActionData":
        """
        Remove a value matching the signature
        :param searched_signature: Looked up signature
        :param ignore_non_existent: Controls whether the method raises an exception if a value with
        the ``signature`` wasn't found
        :param exact_match: Controls whether exact signature should be found
        :return:
        """
        try:
            if exact_match:
                to_remove = self.get_with_signature(searched_signature=searched_signature)[0]
            else:
                to_remove = self.find_with_signature(searched_signature=searched_signature)[0][0]
        except (NothingFound, IndexError):
            if not ignore_non_existent:
                raise
            return self.evolve_self()

        return self.evolve_self(data=tuple(filter(lambda signature_value: signature_value[0] != to_remove, self.data)))

    def remove_many(
        self,
        searched_signatures: Sequence[SignatureT[Any]],
        ignore_non_existent: bool = False,
        exact_match: bool = True,
    ) -> "ActionData":
        """
        Remove values matching the signatures
        :param searched_signatures: Looked up signatures
        :param ignore_non_existent: Controls whether the method raises an exception if a value with
        the ``signature`` wasn't found
        :param exact_match: Controls whether exact signature should be found
        :return: Copy of self
        """

        instance = self
        for searched_signature in searched_signatures:
            instance = instance.remove(
                searched_signature, ignore_non_existent=ignore_non_existent, exact_match=exact_match
            )
        return instance

    def with_new_execution_meta(self) -> "ActionData":
        """
        Clean all metadata from the previous execution. Note that observers will be created from the default classes

        :return: Copy of self with only :attrs:`ActionData.data` preserved
        """
        return ActionData(data=self.data)

    def with_future(self, future: Awaitable[Any]) -> "ActionData":
        """
        Log futures references
        """
        return self.evolve_self(futures=self.futures + [future])

    def record_start(self, action: ActionT) -> "ActionData":
        """
        :param action: Action which is recorded
        :return: New copy with recorded start of the action
        """
        for observer in self.observers:
            if observer.should_record_action(action):
                observer.record_start(action)
        return self

    def record_end(self, action: ActionT) -> "ActionData":
        """
        :param action: Action which is recorded
        :return: New copy with recorded end of the action
        """
        for observer in self.observers:
            if observer.should_record_action(action):
                observer.record_end(action)
        return self

    @classmethod
    def create(cls, **data: Any) -> "ActionData":
        """
        Convenient way how to create the :class:`ActionData` with only key signatures

        :param data: Data values of keys-values
        :return: Created :class:`ActionData`
        """
        return ActionData(data=cls._keyed_dict_as_values_with_signatures(data))

    # Export
    def as_keyed_dict(self) -> Dict[str, Any]:
        """
        Export self into a dictionary for keys-values for signatures which have keys set
        """
        return {signature.key: value for signature, value in self.data if signature.key}

    def with_new_data(self, data: Dict[str, Any]) -> "ActionData":
        """
        :param data: New data to be appended
        :return: Copy with new data
        """
        return self.evolve_self(data=self._keyed_dict_as_values_with_signatures(data))

    @property
    def signatures(self) -> List[SignatureT[Any]]:
        """
        List registered signatures
        """
        return [reg[0] for reg in self.data]

    def is_in(self, key: str) -> bool:
        """
        Find if the key is in the data
        """
        try:
            self.get_by_key(key)
            return True
        except NothingFound:
            return False

    def signature_is_in(self, searched_signature: SignatureT[Any]) -> bool:
        """
        Find if the signature is in the data
        """
        try:
            self.get_by_signature(searched_signature)
            return True
        except NothingFound:
            return False

    def get_observer(self, observer_cls: Type[ObserverVar]) -> ObserverVar:
        """
        :param observer_cls: Class of the observer to find
        :return: Observer from the observers attached to this container
        """
        return self._ensure_one([observer for observer in self.observers if isinstance(observer, observer_cls)])

    @classmethod
    def _get_from_nested(cls, key: str, data: Dict[str, Any], default: Optional[Any] = None) -> Any:
        first_dot_index = key.find(".")
        if first_dot_index > 0:
            if key[:first_dot_index] not in data:
                return default

            return cls._get_from_nested(key[first_dot_index + 1 :], data[key[:first_dot_index]])
        return data.get(key, default)

    @staticmethod
    def _ensure_one(matched: Sequence[T]) -> T:
        n_matched = len(matched)

        if n_matched == 1:
            return matched[0]
        elif n_matched == 0:
            raise NothingFound()

        # Check for duplicates
        first_value = matched[0]
        for value in matched[1:]:
            if value != first_value:
                raise FoundMoreThanOne("Expected one, but found {}".format(n_matched))
        return first_value

    @staticmethod
    def _keyed_dict_as_values_with_signatures(data: Dict[str, Any]) -> Tuple[Tuple[SignatureT[Any], Any], ...]:
        return tuple((Signature(key=key, type_=type(value)), value) for key, value in data.items())



import json
import sys
from abc import ABC, abstractmethod
from functools import wraps
from typing import (
    Iterable,
    Callable,
    Union,
    Type,
    Optional,
    Any,
    Dict,
    ContextManager,
    AsyncContextManager,
    Tuple,
    NoReturn,
    Coroutine,
)

from orinoco import config
from orinoco.entities import ActionData
from orinoco.exceptions import ActionNotProperlyConfigured, BaseActionException
from orinoco.observers import ActionsLog
from orinoco.tags import SystemActionTag
from orinoco.types import ActionT, ActionDataT, NamespacedActionT, ActionVar


class _NamespacedAction(NamespacedActionT):
    """
    Wrapper class named actions. See `Action.namespaced`
    """

    def __init__(self, class_to_return: Type[ActionT]):
        self.class_to_return = class_to_return

    def __getattr__(self, attribute: Any) -> Type[ActionT]:
        self.class_to_return.name = attribute
        return self.class_to_return


class SyncActionMixin(ABC):
    @abstractmethod
    def run(self, action_data: ActionDataT) -> ActionDataT:
        """
        Method which is executed when running the action pipeline

        Note that each override of this function have to implement "skipping logic" to support :class:`~Return`
        functionality.

        .. code-block:: python

            def run(self, action_data: ActionDataT) -> ActionDataT:
                if action_data.skip_processing:
                    return action_data

                # normal implementation of the action
                ...

        :param action_data: Data container passed by actions
        :return: Last version of :class:`:class:~orinoco.entities.ActionData` after execution
        """
        pass

    def run_with_data(self, **params: Any) -> ActionDataT:
        """
        Shortcut for running actions without creating :class:`~orinoco.entities.ActionData` object directly

        :param params: Parameters which are propagated into :class:`~orinoco.entities.ActionData`
        :return: Processed data with results
        """
        return self.run(ActionData.create(**params))


class AsyncActionMixin(ABC):
    @abstractmethod
    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        """
        Method which is executed when running the action pipeline

        Note that each override of this function have to implement "skipping logic" to support :class:`~Return`
        functionality.

        .. code-block:: python

            def run(self, action_data: ActionDataT) -> ActionDataT:
                if action_data.skip_processing:
                    return action_data

                # normal implementation of the action
                ...

        :param action_data: Data container passed by actions
        :return: Last version of `~orinoco.entities.ActionData` after execution
        """
        pass

    async def async_run_with_data(self, **params: Any) -> ActionDataT:
        """
        Shortcut for running actions asynchronously without creating :class:`~orinoco.entities.ActionData` object directly

        :param params: Parameters which are propagated into :class:`~orinoco.entities.ActionData`
        :return: Processed data with results
        """
        return await self.async_run(ActionData.create(**params))


class Action(SyncActionMixin, AsyncActionMixin, ActionT, ABC):
    """
    Fundamental entity which is designed to be chained with other actions so one action process
    the :class:`~orinoco.entities.ActionData` and send it to the next action.
    """

    DESCRIPTION: str = ""
    NAME: str = ""

    def __init__(self, description: Optional[str] = None, name: Optional[str] = None):
        """
        :param description: Description of the task. Is used when an exception is raised, for debugging and
        increase of readability. It has no effect for runtime. Can be also set as class attribute if it's
        not provided to constructor
        """
        self.description = description or self.DESCRIPTION
        self.name = name or self.NAME or self.__class__.__name__

    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        return self.run(action_data)

    def then(self, another_action: "ActionT") -> "ActionT":
        """
        One of the chaining functions. More convenient way for chaining is to use `>>` operator like
        this:

        .. code-block:: python

            Action1() >> Action2()

        :param another_action: Another action which will be executed after
        :return: Set of this action and ``another_action`` represented by :class:`ActionSet`
        """
        return ActionSet([self, another_action])

    def __rshift__(self, other: ActionT) -> "ActionT":
        return self.then(other)

    def on_subfield(self, field_key: str) -> "OnActionDataSubField":
        """
        Run action on a subfield. See :class:`OnActionDataSubField` class for more information.

        :param field_key: Key to the item in :class:`~orinoco.entities.ActionData` on which the action will be executed
        :return: Wrapped action which will be executed on the subfield
        """
        return OnActionDataSubField(self, field_key)

    @property
    def action_name(self) -> str:
        """
        :return: Name of the action
        """
        return self.name or self.__class__.__name__

    @classmethod
    def namespaced(cls) -> NamespacedActionT:
        """
        Return 'named' instance of itself. It has no effect, just make a 'title' for the action/s by wrapping them
        into a named group of actions.


        Example:

        .. code-block:: python

            EventSet.namespaced().MarkPaymentAttemptAsProcessed(
                    [SetPaymentAuthorizationAttemptProcessed(), SaveModel("authorization_attempt")]
                )

        `MarkPaymentAttemptAsProcessed` is just a "made up" key of the following actions.
        """
        return _NamespacedAction(cls)

    def finish(self) -> "Return":
        return Return(self)


def record_action(
    fu: Callable[[ActionVar, ActionDataT], ActionDataT]
) -> Callable[[ActionVar, ActionDataT], ActionDataT]:
    """
    Decorator for :func:`~SyncActionMixin.run` which records decorated action signature to
    the :class:`~orinoco.entities.ActionData`

    :param fu: :func:`~SyncActionMixin.run` method (or any with the same signature)
    :return: Decorated function
    """

    @wraps(fu)
    def wrapper(action: ActionVar, action_data: ActionDataT) -> ActionDataT:
        if config.CHAINING_TYPE_CHECK_STRICT_MODE_ENABLED and not isinstance(action_data, ActionData):
            raise ActionNotProperlyConfigured(
                "Input argument of {} - {} must of type ActionData, but got {}".format(
                    action.action_name, fu, action_data
                )
            )

        result = fu(action, action_data.record_start(action))

        if config.CHAINING_TYPE_CHECK_STRICT_MODE_ENABLED and not isinstance(result, ActionData):
            raise ActionNotProperlyConfigured(
                "Output of {} - {} must of type ActionData, but got {}".format(action.action_name, fu, result)
            )

        return result.record_end(action)

    return wrapper


def async_record_action(
    fu: Callable[[ActionVar, ActionDataT], Coroutine[None, None, ActionDataT]]
) -> Callable[[ActionVar, ActionDataT], Coroutine[None, None, ActionDataT]]:
    """
    Decorator for :func:`~AsyncActionMixin.async_run` which records decorated action signature to
    the :class:`~orinoco.entities.ActionData`

    :param fu: :func:`~AsyncActionMixin.async_run` method (or any with the same signature)
    :return: Decorated function
    """

    @wraps(fu)
    async def wrapper(action: ActionVar, action_data: ActionDataT) -> ActionDataT:
        if config.CHAINING_TYPE_CHECK_STRICT_MODE_ENABLED and not isinstance(action_data, ActionData):
            raise ActionNotProperlyConfigured(
                "Input argument of {} - {} must of type ActionData, but got {}".format(
                    action.action_name, fu, action_data
                )
            )

        result = await fu(action, action_data.record_start(action))

        if config.CHAINING_TYPE_CHECK_STRICT_MODE_ENABLED and not isinstance(result, ActionData):
            raise ActionNotProperlyConfigured(
                "Output of {} - {} must of type ActionData, but got {}".format(action.action_name, fu, result)
            )

        return result.record_end(action)

    return wrapper


def verbose_action_exception(
    fu: Callable[[ActionVar, ActionDataT], ActionDataT]
) -> Callable[[ActionVar, ActionDataT], ActionDataT]:
    """
    Decorator which adds some more context information to errors raised during action execution.

    :param fu: :func:`~SyncActionMixin.run` method (or any with the same signature)
    :return: Decorated function
    """

    @wraps(fu)
    def wrapper(action: ActionVar, action_data: ActionDataT) -> ActionDataT:
        try:
            return fu(action, action_data)
        except Exception as err:
            _raise_new_error(action=action, err=err, action_data=action_data)

    return wrapper


def async_verbose_action_exception(
    fu: Callable[[ActionVar, ActionDataT], Coroutine[None, None, ActionDataT]]
) -> Callable[[ActionVar, ActionDataT], Coroutine[None, None, ActionDataT]]:
    """
     Decorator which adds some more context information to errors raised during action execution.

    :param fu: :func:`~AsyncActionMixin.async_run` method (or any with the same signature)
    :return: Decorated function
    """

    @wraps(fu)
    async def wrapper(action: ActionVar, action_data: ActionDataT) -> ActionDataT:
        try:
            return await fu(action, action_data)
        except Exception as err:
            _raise_new_error(action=action, err=err, action_data=action_data)

    return wrapper


def _raise_new_error(action: ActionT, err: Exception, action_data: ActionDataT) -> NoReturn:
    if config.VERBOSE_ERRORS and (len(err.args) == 0 or (err.args and "Action context" not in err.args[0])):
        fields = [
            ("Actions history", json.dumps(action_data.get_observer(ActionsLog).actions_log, indent=4)),
            ("Actions data", json.dumps({str(k): str(v) for k, v in action_data.data}, indent=4)),
            (
                "{} params".format(action.action_name),
                json.dumps({k: str(v) for k, v in action.__dict__.items()}, indent=4),
            ),
        ]

        context_message = "\n".join("{}: {}".format(field[0], field[1]) for field in fields)

        msg = "{0}\n{1} Action context {1}\n{2}".format(str(err).strip(), "-" * 20, context_message)
        try:
            raise err.__class__(msg).with_traceback(sys.exc_info()[2]) from err
        except TypeError:
            raise BaseActionException(msg)
    else:
        raise


class ActionSet(Action, SystemActionTag):
    """
    Set of actions which are executed consequently
    """

    ACTIONS: Optional[Iterable[ActionT]] = None

    def __init__(
        self, actions: Optional[Iterable[ActionT]] = None, description: Optional[str] = None, name: Optional[str] = None
    ):
        super().__init__(description=description, name=name)
        self.actions = actions or self.ACTIONS or []

    @record_action
    @verbose_action_exception
    def run(self, action_data: ActionDataT) -> ActionDataT:
        for action in self.actions:
            if action_data.skip_processing:
                return action_data

            action_data = action.run(action_data)
        return action_data

    @async_record_action
    @async_verbose_action_exception
    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        for action in self.actions:
            if action_data.skip_processing:
                return action_data

            action_data = await action.async_run(action_data)
        return action_data


class AtomicActionSet(ActionSet, SystemActionTag):
    """
    Set of actions which are executed in a context manager
    """

    def __init__(
        self,
        actions: Iterable[ActionT],
        atomic_context_manager: Callable[[], ContextManager[None]],
        description: Optional[str] = None,
        name: Optional[str] = None,
    ):
        super().__init__(actions=actions, description=description, name=name)
        self.atomic_context_manager = atomic_context_manager

    @record_action
    @verbose_action_exception
    def run(self, action_data: ActionDataT) -> ActionDataT:
        with self.atomic_context_manager():
            for action in self.actions:
                if action_data.skip_processing:
                    return action_data

                action_data = action.run(action_data)
            return action_data


class AsyncAtomicActionSet(ActionSet, SystemActionTag):
    """
    Async version of :class:`AtomicActionSet`.
    """

    def __init__(
        self,
        actions: Iterable[ActionT],
        atomic_context_manager: Callable[[], AsyncContextManager[None]],
        description: Optional[str] = None,
        name: Optional[str] = None,
    ):
        super().__init__(actions=actions, description=description, name=name)
        self.atomic_context_manager = atomic_context_manager

    @async_record_action
    @async_verbose_action_exception
    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        async with self.atomic_context_manager():
            for action in self.actions:
                if action_data.skip_processing:
                    return action_data

                action_data = await action.async_run(action_data)
            return action_data


class Then(ActionSet, SystemActionTag):
    """
    Syntactics sugar for :class:`ActionSet`. See :class:`~orinoco.condition.Switch` docs for usage
    """

    def __init__(self, *actions: ActionT):
        super().__init__(actions)


class HandledExceptions(ActionSet, SystemActionTag):
    """
    Try-except wrapper for actions
    """

    def __init__(
        self,
        *actions: ActionT,
        catch_exceptions: Union[Tuple[Type[BaseException], ...], Type[BaseException]],
        handle_method: Callable[[BaseException, ActionDataT], Optional[Union[Type[BaseException], BaseException]]],
        fail_on_error: bool = True,
        name: Optional[str] = None,
    ):
        """
        :param actions: Actions to execute
        :param catch_exceptions: Errors to except
        :param handle_method: Method to execute when the actions fail
        :param fail_on_error: If true fail on error, otherwise the execution will continue
        :param name: Name of the action
        """

        super().__init__(actions=actions, name=name)
        self.catch_exceptions = catch_exceptions
        self.handle_method = handle_method
        self.fail_on_error = fail_on_error

    def run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data
        try:
            return super().run(action_data)
        except self.catch_exceptions as err:
            new_err = self.handle_method(err, action_data)
            if not self.fail_on_error:
                return action_data
            if isinstance(new_err, BaseException):
                raise new_err
            raise err

    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        try:
            return await super(ActionSet, self).async_run(action_data=action_data)
        except self.catch_exceptions as err:
            new_err = self.handle_method(err, action_data)

            if not self.fail_on_error:
                return action_data
            if isinstance(new_err, BaseException):
                raise new_err
            raise err


class OnActionDataSubField(Action, SystemActionTag):
    """
    Utility action to execute an action on values of a nested item.

    For example if an action accepts parameter `user`, but action data contains only `request_data` dictionary with
    the desired `user`.

    .. code-block:: python

        action = OnActionDataSubField(DoSomethingWithUser(), field_key="request_data")

        # Same as
        action = DoSomethingWithUser().on_subfield("request_data")

    As shown above :func:`~Action.on_subfield` can be used as a shortcut.
    """

    def __init__(self, action: ActionT, field_key: str, name: Optional[str] = None):
        super().__init__(name=name or "{}[{}({})]".format(self.__class__.__name__, action.action_name, field_key))
        self.action = action
        self.field_key = field_key

    def run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        return self._with_subfield_data(
            action_data, self.action.run(action_data.with_new_data(action_data.get(self.field_key)))
        )

    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        return self._with_subfield_data(
            action_data, await self.action.async_run(action_data.with_new_data(action_data.get(self.field_key)))
        )

    def _with_subfield_data(self, action_data: ActionDataT, new_action_data: ActionDataT) -> ActionDataT:
        return action_data.evolve(**self._unwrap_nested_key(self.field_key, new_action_data.as_keyed_dict()))

    @classmethod
    def _unwrap_nested_key(cls, key: str, value: Any) -> Dict[str, Any]:
        last_dot_index = key.rfind(".")
        if last_dot_index > 0:
            return cls._unwrap_nested_key(key[:last_dot_index], value={key[last_dot_index + 1 :]: value})

        return {key: value}


class Return(Action):
    """
    Mark action data as "should not be processed", imitating return statement in functions
    """

    def __init__(self, action: Optional[ActionT] = None):
        super().__init__()
        self.action = action

    def run(self, action_data: ActionDataT) -> ActionDataT:
        return (self.action.run(action_data) if self.action else action_data).evolve_self(skip_processing=True)



import asyncio
from abc import ABC
from typing import Iterable, Callable, Optional

from orinoco.action import (
    Action,
    record_action,
    verbose_action_exception,
    ActionSet,
    async_record_action,
    async_verbose_action_exception,
)
from orinoco.exceptions import ActionNotProperlyInherited
from orinoco.tags import SystemActionTag
from orinoco.types import ActionDataT


class Event(Action, ABC):
    """
    Base class for event based actions which run a side-effect
    """

    def __init__(self, description: Optional[str] = None, async_blocking: bool = False, name: Optional[str] = None):
        """
        :param async_blocking: Controls whether to wait for the result when running asynchronously
        """
        self.async_blocking = async_blocking
        super().__init__(description=description, name=name)

    @record_action
    @verbose_action_exception
    def run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        self.run_side_effect(action_data)
        return action_data

    @async_record_action
    @async_verbose_action_exception
    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        coroutine = self.async_run_side_effect(action_data)
        if self.async_blocking:
            await coroutine
            return action_data
        return action_data.with_future(asyncio.create_task(coroutine))

    def run_side_effect(self, action_data: ActionDataT) -> None:
        raise ActionNotProperlyInherited(
            "`run_side_effect` method needs to implemented for sync execution or `async_run_side_effect` "
            "for async execution"
        )

    async def async_run_side_effect(self, action_data: ActionDataT) -> None:
        self.run_side_effect(action_data)


class GenericEvent(Event):
    """
    Utility action which allows running side-effects "on the fly" (usually by lambda functions)
    """

    def __init__(
        self,
        method: Callable[[ActionDataT], None],
        description: Optional[str] = None,
        async_blocking: bool = False,
        name: Optional[str] = None,
    ):
        super().__init__(description, async_blocking=async_blocking, name=name)
        self.method = method

    def run_side_effect(self, action_data: ActionDataT) -> None:
        self.method(action_data)


class EventSet(Event, SystemActionTag):
    """
    Isolated ActionSet - none of the ActionData modification is propagated further
    """

    def __init__(
        self,
        actions: Iterable[Action],
        description: Optional[str] = None,
        async_blocking: bool = False,
        name: Optional[str] = None,
    ):
        super().__init__(description=description, async_blocking=async_blocking, name=name)
        self.actions_set = ActionSet(actions=actions, name=name)

    def run_side_effect(self, action_data: ActionDataT) -> None:
        self.actions_set.run(action_data)



import string
from typing import Iterable, Type, List, Any, NoReturn, Union, Optional, Tuple, Set

from typing_extensions import Annotated, get_origin, get_args

from orinoco.types import TypeT, AnnotationNameT


def initialize(classes: Iterable[Type[Any]]) -> List[Any]:
    return [cls_object() for cls_object in classes]


def raise_not_provided_field(field_name: str) -> NoReturn:
    raise ValueError("Field {} has to be provided".format(field_name))


def is_format_string(value: str) -> bool:
    return any(get_format_string_args(value))


def get_format_string_args(value: str) -> List[str]:
    return [tup[1] for tup in string.Formatter().parse(value) if (tup[1] is not None and tup[1] != "")]


def extract_type(
    value: Union[TypeT, Annotated[TypeT, AnnotationNameT]]
) -> Tuple[TypeT, Optional[AnnotationNameT], Set[str]]:
    generic = get_origin(value)
    if generic is Annotated:
        dtype, annotation, *tags = get_args(value)  # first two parameters - (type, annotation)
        return dtype, annotation, set(tags)
    return value, None, set()



from abc import ABC, abstractmethod
from typing import Dict, Any, TypeVar, Optional, Type, List, Tuple, Sequence, ClassVar, Set, Generic

from pydantic import BaseModel


T = TypeVar("T")
BaseModelT = TypeVar("BaseModelT", bound=BaseModel)


class NamespacedActionT:
    class_to_return: Type["ActionT"]

    @abstractmethod
    def __getattr__(self, _: Any) -> Type["ActionT"]:
        pass


class ImmutableEvolvableModelT(BaseModel, ABC):
    class Config:
        allow_mutation = False
        arbitrary_types_allowed = True

    @abstractmethod
    def evolve_self(self: T, **kwargs: Any) -> T:
        pass


class SignatureT(Generic[T], ImmutableEvolvableModelT, ABC):
    type_: Optional[Type[T]]
    tags: Set[str]
    key: Optional[str]

    @abstractmethod
    def match(self, other_signature: "SignatureT[T]") -> bool:
        pass


class ActionConfigT(Generic[T], ImmutableEvolvableModelT, ABC):
    INPUT: Optional[Dict[str, SignatureT[Any]]]
    OUTPUT: Optional[SignatureT[T]]


class ObserverT(ABC):
    @abstractmethod
    def should_record_action(self, action: "ActionT") -> bool:
        pass

    @abstractmethod
    def record_start(self, action: "ActionT") -> None:
        pass

    @abstractmethod
    def record_end(self, action: "ActionT") -> None:
        pass


class ActionDataT(ImmutableEvolvableModelT, ABC):

    NOT_FOUND: ClassVar = object()

    # data: List[Tuple[SignatureT[Any], Any]]
    data: Tuple[Tuple[SignatureT[Any], Any], ...]
    futures: List[Any]
    observers: List[ObserverT]
    skip_processing: bool

    @abstractmethod
    # @property
    def signatures(self) -> List[SignatureT[Any]]:
        pass

    @abstractmethod
    def __getitem__(self, searched_signature: SignatureT[T]) -> T:
        pass

    @abstractmethod
    def get(self, name: str, default: Any = NOT_FOUND) -> Any:
        pass

    @abstractmethod
    def get_by_signature(self, searched_signature: SignatureT[T]) -> T:
        pass

    @abstractmethod
    def get_with_signature(self, searched_signature: SignatureT[T]) -> Tuple[SignatureT[T], T]:
        pass

    @abstractmethod
    def get_or_default(self, key: str, default: Optional[Any] = None) -> Any:
        pass

    @abstractmethod
    def find_or_default(self, key: str, default: Optional[Any] = None) -> Any:
        pass

    @abstractmethod
    def get_by_key(self, name: str, default: Any = NOT_FOUND) -> Any:
        pass

    @abstractmethod
    def get_by_type(self, type_: Type[T]) -> T:
        pass

    @abstractmethod
    def get_by_tags(self, *tags: str) -> Any:
        pass

    @abstractmethod
    def find(self, searched_signature: SignatureT[T]) -> List[T]:
        pass

    @abstractmethod
    def find_one(self, searched_signature: SignatureT[T]) -> T:
        pass

    @abstractmethod
    def find_with_signature(self, searched_signature: SignatureT[T]) -> List[Tuple[SignatureT[T], T]]:
        pass

    @abstractmethod
    def evolve(self, **data: Any) -> "ActionDataT":
        pass

    @abstractmethod
    def register(self, signature: SignatureT[T], entity: T, check_if_exists: bool = True) -> "ActionDataT":
        pass

    @abstractmethod
    def register_many(self, data: Sequence[Tuple[SignatureT[T], T]], check_if_exists: bool = True) -> "ActionDataT":
        pass

    @abstractmethod
    def remove(
        self, searched_signature: SignatureT[Any], ignore_non_existent: bool = False, exact_match: bool = True
    ) -> "ActionDataT":
        pass

    @abstractmethod
    def remove_many(
        self,
        searched_signatures: Sequence[SignatureT[Any]],
        ignore_non_existent: bool = False,
        exact_match: bool = True,
    ) -> "ActionDataT":
        pass

    @abstractmethod
    def with_future(self, future: Any) -> "ActionDataT":
        pass

    @abstractmethod
    def record_start(self, action: "ActionT") -> "ActionDataT":
        pass

    @abstractmethod
    def record_end(self, action: "ActionT") -> "ActionDataT":
        pass

    @classmethod
    @abstractmethod
    def create(cls, **data: Any) -> "ActionDataT":
        pass

    @abstractmethod
    def as_keyed_dict(self) -> Dict[str, Any]:
        pass

    @abstractmethod
    def with_new_data(self, data: Dict[str, Any]) -> "ActionDataT":
        pass

    @abstractmethod
    def is_in(self, key: str) -> bool:
        pass

    @abstractmethod
    def signature_is_in(self, searched_signature: SignatureT[Any]) -> bool:
        pass

    @abstractmethod
    def get_observer(self, observer_cls: Type["ObserverVar"]) -> "ObserverVar":
        pass

    @abstractmethod
    def with_new_execution_meta(self) -> "ActionDataT":
        pass


class ActionT(ABC):

    DESCRIPTION: str
    NAME: str

    description: str
    name: str

    # @abstractmethod
    @property
    def action_name(self) -> str:
        pass

    @abstractmethod
    def then(self, another_action: "ActionT") -> "ActionT":
        pass

    @abstractmethod
    def __rshift__(self, other: "ActionT") -> "ActionT":
        pass

    @abstractmethod
    def on_subfield(self, field_key: str) -> "ActionT":
        pass

    @classmethod
    @abstractmethod
    def namespaced(cls) -> NamespacedActionT:
        pass

    @abstractmethod
    def run(self, action_data: ActionDataT) -> ActionDataT:
        pass

    @abstractmethod
    def run_with_data(self, **params: Any) -> ActionDataT:
        pass

    @abstractmethod
    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        pass

    @abstractmethod
    async def async_run_with_data(self, **params: Any) -> ActionDataT:
        pass

    @abstractmethod
    def finish(self) -> "ActionT":
        pass


ActionVar = TypeVar("ActionVar", bound=ActionT)
ObserverVar = TypeVar("ObserverVar", bound=ObserverT)

TypeT = TypeVar("TypeT")
AnnotationNameT = TypeVar("AnnotationNameT", bound=str)


ErrorT = TypeVar("ErrorT", bound=BaseException)



from typing import Optional


class BaseActionException(Exception):
    def __init__(self, message: str = "", reason: Optional[str] = None, action: Optional[object] = None):
        super().__init__(message)
        self.reason = reason
        self.action = action
        self.message = message

    def __str__(self) -> str:
        if self.reason is not None and self.action is not None:
            return "{}: {}: {}: {}".format(
                self.__class__.__name__, self.action.__class__.__name__, self.reason, self.message
            )
        return "{}: {}".format(self.__class__.__name__, self.message)


class ConditionNotMet(BaseActionException):
    pass


class NoneOfActionsCanBeExecuted(BaseActionException):
    pass


class KeyNotInActionData(BaseActionException):
    pass


class ActionNotReturnedActionData(BaseActionException):
    pass


class ActionNotProperlyInherited(RuntimeError):
    pass


class ActionNotProperlyConfigured(RuntimeError):
    pass


##
class SearchError(BaseActionException):
    pass


class NothingFound(SearchError):
    pass


class FoundMoreThanOne(SearchError):
    pass


class AlreadyRegistered(BaseActionException):
    pass


class RunnableOnlyInAsyncContext(BaseActionException):
    pass


class RetryError(BaseActionException):
    pass



import abc
import time
from typing import List, Tuple, Dict

from orinoco.tags import SystemActionTag
from orinoco.types import ObserverT, ActionT


class Observer(ObserverT, abc.ABC):
    def should_record_action(self, action: ActionT) -> bool:
        """
        Controls whether the observer will log given action
        """
        return True

    def record_start(self, action: ActionT) -> None:
        """
        This method is executed before the ``action`` is executed
        """
        pass

    def record_end(self, action: ActionT) -> None:
        """
        This method is executed after the ``action`` is executed
        """
        pass


class ExecutionTimeObserver(Observer):
    """
    Observer which measures execution times of the actions in the pipeline
    """

    measurements: List[Tuple[str, float]]

    def __init__(self) -> None:
        self.measurements = []
        self._measuring: Dict[ActionT, float] = {}

    def __repr__(self) -> str:
        return "ExecutionTimeObserver({})".format(self.measurements)

    def should_record_action(self, action: ActionT) -> bool:
        return SystemActionTag not in action.__class__.__bases__

    def record_start(self, action: ActionT) -> None:
        self._measuring[action] = time.time()

    def record_end(self, action: ActionT) -> None:
        self.measurements.append(("action.{}".format(action.action_name), time.time() - self._measuring.pop(action)))


class ActionsLog(Observer):
    """
    Observer which logs which actions were executed
    """

    actions_log: List[str]

    def __init__(self) -> None:
        self.actions_log = []

    def __repr__(self) -> str:
        return "ActionsLog({})".format(self.actions_log)

    def record_start(self, action: ActionT) -> None:
        self.actions_log.append(
            (
                action.action_name + "_start"
                if "AND" not in action.action_name
                else "({})_start".format(action.action_name)
            )
        )

    def record_end(self, action: ActionT) -> None:
        self.actions_log.append(
            action.action_name + "_end" if "AND" not in action.action_name else "({})_end".format(action.action_name)
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['orinoco']

package_data = \
{'': ['*']}

install_requires = \
['mypy>=0.991,<0.992', 'pydantic>=1.7.3,<2.0.0', 'returns>=0.19.0,<0.20.0']

setup_kwargs = {
    'name': 'orinoco',
    'version': '1.0.8',
    'description': 'Functional composable pipelines allowing clean separation of the business logic and its implementation',
    'long_description': 'None',
    'author': 'Martin Vo',
    'author_email': 'mvo@paysure.solutions',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import abc
import time
from datetime import datetime
from enum import Enum, auto
from typing import Any, Callable, Type, Generic, Optional, cast

from pydantic import BaseModel, Field
from returns import pipeline

from returns.result import Failure, Result, Success

from orinoco.action import record_action, verbose_action_exception, Action
from orinoco.exceptions import RetryError, ConditionNotMet, BaseActionException
from orinoco.types import ActionT, ActionDataT, ErrorT


class RetryStatus(Enum):
    STARTED = auto()
    SUCCESSFUL = auto()
    FAILED = auto()


class RetryInfo(BaseModel):
    max_retries: int
    retry_delay: float
    retry_count: int = 0
    status: RetryStatus = RetryStatus.STARTED

    started: datetime = Field(default_factory=datetime.now)
    finished: Optional[datetime] = None

    class Config:
        arbitrary_types_allowed = False
        allow_mutation = False


class AbstractRetry(Generic[ErrorT], Action, abc.ABC):
    def __init__(self, action: ActionT, max_retries: int = 10, retry_delay: float = 10):
        super().__init__(name=f"<{self.__class__.__name__}: {action.name} every {retry_delay}s (max {max_retries})>")
        self.action = action
        self.max_retries = max_retries
        self.retry_delay = retry_delay

    @record_action
    @verbose_action_exception
    def run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        retry_info = RetryInfo(max_retries=self.max_retries, retry_delay=self.retry_delay)
        for i in range(self.max_retries):
            result_action_data_result = self._run_action(action_data)

            is_successful = self._is_successful(result_action_data_result)
            retry_info = self._get_new_retry_info(is_successful, i + 1, retry_info)
            if is_successful:
                result_action_data = result_action_data_result.unwrap()
                return result_action_data.evolve(
                    retry_infos=result_action_data.get("retry_infos", []) + [(self.action.name, retry_info)]
                )

            time.sleep(self.retry_delay)
        raise RetryError("{} failed after {} retries. Info: {}".format(self.name, self.max_retries, retry_info))

    def _get_new_retry_info(self, is_successful: bool, retry_counter: int, previous_retry_info: RetryInfo) -> RetryInfo:
        retry_status = self._get_retry_status(is_successful, retry_counter)
        retry_info = previous_retry_info.copy(update={"status": retry_status, "retry_count": retry_counter})
        if retry_status is RetryStatus.SUCCESSFUL:
            return retry_info.copy(update={"finished": datetime.utcnow()})

        return retry_info

    def _get_retry_status(self, is_successful: bool, retry_counter: int) -> RetryStatus:
        if is_successful:
            return RetryStatus.SUCCESSFUL
        elif retry_counter > self.max_retries:
            return RetryStatus.FAILED
        return RetryStatus.STARTED

    def _run_action(self, action_data: ActionDataT) -> Result[ActionDataT, ErrorT]:
        try:
            return Success(self.action.run(action_data))
        except ConditionNotMet as err:
            return Failure(cast(ErrorT, err))
        except BaseActionException:
            raise
        except BaseException as err:
            return Failure(cast(ErrorT, err))

    @abc.abstractmethod
    def _is_successful(self, value: Result[ActionDataT, ErrorT]) -> bool:
        pass


class AbstractNotFailingRetryWithKey(Generic[ErrorT], AbstractRetry[ErrorT], abc.ABC):
    def __init__(self, action: ActionT, key: str, max_retries: int = 10, retry_delay: float = 10):
        super().__init__(action=action, max_retries=max_retries, retry_delay=retry_delay)
        self.key = key

    def _is_successful(self, value: Result[ActionDataT, ErrorT]) -> bool:
        return self._check_condition(value.unwrap().get(self.key))

    @abc.abstractmethod
    def _check_condition(self, value: Any) -> bool:
        pass


class AbstractFailingRetry(Generic[ErrorT], AbstractRetry[ErrorT], abc.ABC):
    def _is_successful(self, action_data_result: Result[ActionDataT, ErrorT]) -> bool:
        if pipeline.is_successful(action_data_result):
            return True
        if self._check_exception(action_data_result.failure()):
            return False
        raise action_data_result.failure()

    @abc.abstractmethod
    def _check_exception(self, exception: BaseException) -> bool:
        pass


class WaitUntilTrue(AbstractFailingRetry[ConditionNotMet]):
    def _check_exception(self, value: BaseException) -> bool:
        return isinstance(value, ConditionNotMet)


class WaitUntilEqualsTo(AbstractNotFailingRetryWithKey):
    def __init__(self, action: ActionT, key: str, value: Any, max_retries: int = 10, retry_delay: float = 10):
        super().__init__(action=action, key=key, max_retries=max_retries, retry_delay=retry_delay)
        self.value = value

    def _check_condition(self, value: Any) -> bool:
        return value == self.value


class WaitUntilContains(AbstractNotFailingRetryWithKey):
    def __init__(self, action: ActionT, key: str, value: Any, max_retries: int = 10, retry_delay: float = 10):
        super().__init__(action=action, key=key, max_retries=max_retries, retry_delay=retry_delay)
        self.value = value

    def _check_condition(self, value: Any) -> bool:
        return self.value in value


class WaitUntilNotFail(Generic[ErrorT], AbstractFailingRetry[ErrorT]):
    def __init__(
        self,
        action: ActionT,
        exception_cls_to_catch: Optional[Type[ErrorT]] = None,
        max_retries: int = 10,
        retry_delay: float = 10,
    ):
        super().__init__(action=action, max_retries=max_retries, retry_delay=retry_delay)
        self.exception_cls_to_catch = exception_cls_to_catch

    def _check_exception(self, exception: BaseException) -> bool:
        return isinstance(exception, self.exception_cls_to_catch) if self.exception_cls_to_catch else True


class WaitForGeneric(AbstractNotFailingRetryWithKey):
    def __init__(
        self,
        action: ActionT,
        key: str,
        check_condition_function: Callable[[Any], bool],
        max_retries: int = 10,
        retry_delay: int = 10,
    ):
        super().__init__(action=action, key=key, max_retries=max_retries, retry_delay=retry_delay)
        self.check_condition_function = check_condition_function

    def _check_condition(self, value: Any) -> bool:
        return self.check_condition_function(value)



import copy
import json
from abc import abstractmethod, ABC
from typing import Tuple, Callable, Any, Iterable, List, Generator, Union, Optional, Type, NoReturn, TypeVar, Generic

from orinoco.action import (
    Action,
    record_action,
    verbose_action_exception,
    Then,
    async_record_action,
    async_verbose_action_exception,
)
from orinoco.entities import ActionData, Signature
from orinoco.exceptions import NoneOfActionsCanBeExecuted, ConditionNotMet
from orinoco.helpers import raise_not_provided_field, is_format_string, get_format_string_args
from orinoco.tags import SystemActionTag
from orinoco.types import T, ActionDataT, ActionT

ConditionT = TypeVar("ConditionT", bound="Condition")


class Condition(Action, ABC):
    """
    Base class for condition base actions. In the action set they are treated as checks which raise an exception
    when the condition is not met. However, can be used just for evaluation - see `Switch`.
    """

    FAIL_MESSAGE: str = ""
    ERROR_CLS: Type[Exception] = ConditionNotMet
    INVERTED_COND_PREFIX: str = "Inverted condition: "
    DEFAULT_FORMAT_VALUE = "<NOT-PROVIDED>"

    def __init__(
        self,
        fail_message: Optional[str] = None,
        is_inverted: bool = False,
        error_cls: Optional[Type[Exception]] = None,
        name: Optional[str] = None,
    ):
        """
        :param fail_message: Message which is raised when condition is not met
        :param is_inverted: Whether condition is inverted
        :param error_cls: Exception class which will be raised if provided
        """
        self.fail_message = fail_message or self.FAIL_MESSAGE
        self.is_inverted = is_inverted
        self.error_cls = error_cls or self.ERROR_CLS
        super().__init__(description=self.fail_message, name=name)

    @record_action
    @verbose_action_exception
    def run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        if self.validate(action_data):
            return action_data
        self.fail(action_data)

    def and_(self, condition: "Condition") -> ActionT:
        """
        Syntactic sugar for :func:`~orinoco.action.Action.then` - make more sense for conditions

        :param condition: Another condition which is evaluated after
        :return: :class:`~orinoco.action.ActionSet` of this condition and ``condition``
        """
        return self.then(condition)

    def fail(self, action_data: ActionDataT) -> NoReturn:
        """
        Method which is called when the condition is not met (and called by `Condition.run`)

        :param action_data: Action data for extracting custom information for the exception
        :return:
        """
        fail_message = self.fail_message
        fail_message_format_args = get_format_string_args(fail_message)
        if fail_message_format_args:
            format_dict = {
                k: action_data.get_by_key(k, default=self.DEFAULT_FORMAT_VALUE) for k in fail_message_format_args
            }
            print(format_dict)
            fail_message = fail_message.format(**format_dict)
        raise self.error_cls(
            "{}{} failed: {}".format("not " if self.is_inverted else "", self.__class__.__name__, fail_message)
        )

    def validate(self, action_data: ActionDataT) -> bool:
        """
        Main method which is used for evaluation

        :param action_data: Data container to be evaluated
        :return: Whether condition is met
        """
        val = self._is_valid(action_data)
        if self.is_inverted:
            return not val
        return val

    def validate_with(self, **params: Any) -> bool:
        """
        Shortcut for running validation without creating :class:`~orinoco.entities.ActionData` object directly

        :param params: Parameters which are propagated into action data
        :return: Whether condition is met
        """
        return self.validate(ActionData.create(**params))

    @abstractmethod
    def _is_valid(self, action_data: ActionDataT) -> bool:
        """
        Only abstract method which needs to be implemented by conditions. Please not that it's not intended
        for direct use since it's doesn't consider if the condition is inverted

        :param action_data: Data container to be evaluated
        :return: Whether condition is met
        """
        pass

    def __and__(self, other: "Condition") -> "AndOperator":
        """
        See :class:`AndOperator` docs. It provides and operator: `condition1 & condition2`

        :param other: Condition
        :return: :class:`AndOperator` instance (set of actions)
        """
        return AndOperator(self, other)

    def __or__(self, other: "Condition") -> "OrOperator":
        """
        See :class:`OrOperator` docs. It provides `or` operator: `condition1 | condition2`

        :param other: Second condition
        :return: :class:`OrOperator` instance (set of actions)
        """
        return OrOperator(self, other)

    def __invert__(self: ConditionT) -> ConditionT:
        return self.not_()

    def not_(self: ConditionT) -> ConditionT:
        """
        :return: Inverted condition
        """
        inv_cond = copy.deepcopy(self)
        inv_cond.is_inverted = not self.is_inverted
        inv_cond.fail_message = (
            "{}{}".format(self.INVERTED_COND_PREFIX, inv_cond.fail_message)
            if not inv_cond.fail_message.startswith(self.INVERTED_COND_PREFIX)
            else inv_cond.fail_message[len(self.INVERTED_COND_PREFIX) :]
        )
        inv_cond.description = inv_cond.fail_message
        return inv_cond

    def if_then(self, action: ActionT) -> "ConditionalAction":
        """
        Execute given action when this condition is valid

        :param action: Action to execute
        :return: :class:`ConditionalAction`
        """
        return ConditionalAction(self, action)

    @property
    def name_with_inverted(self) -> str:
        """
        :return: Name of the condition with inverted prefix if needed
        """
        return f'{"~" if self.is_inverted else ""}{self.name}'


class AlwaysTrue(Condition):
    """
    Dummy condition which is always true
    """

    def _is_valid(self, action_data: ActionDataT) -> bool:
        return True


class GenericCondition(Condition):
    """
    Utility condition which allows creating conditions "on the fly" (usually by lambda functions)
    """

    def __init__(
        self,
        validation_method: Callable[[ActionDataT], bool],
        fail_message: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        :param validation_method: Method which returns whether the condition is met
        :param fail_message: Message which will be included in the raise exception
        """
        super().__init__(fail_message=fail_message, name=name)
        self.validation_method = validation_method

    def _is_valid(self, action_data: ActionDataT) -> bool:
        return self.validation_method(action_data)


class PropertyCondition(Condition):
    """
    Generic condition which evaluates property of an object in the dataset
    """

    PROPERTY_OBJECT: Optional[str] = None
    ATTRIBUTE: Optional[str] = None
    EQUAL_TO: Optional[Any] = True

    SOMETHING: Any = object

    attribute: str

    def __init__(
        self,
        property_object: Optional[str] = None,
        attribute: Optional[str] = None,
        equal_to: Any = None,
        fail_message: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Note that when parameters of constructor are not provided, class attributes are used. This could
        be convenient for subclassing without need to override constructor.

        :param property_object: Key of the object in the `ActionData`
        :param attribute: Attribute of the object to be evaluated
        :param equal_to: Expected value of the object's attribute
        :param fail_message:
        """
        super().__init__(
            fail_message=fail_message,
            name=name
            or "{}[{} {} {}]".format(self.__class__.__name__, attribute, "==" if equal_to else "!=", property_object),
        )
        self.attribute = attribute or self.ATTRIBUTE or raise_not_provided_field("attribute")
        self.equal_to = equal_to if equal_to is not None else self.EQUAL_TO
        self.property_object = property_object or self.PROPERTY_OBJECT or raise_not_provided_field("property_object")

    def _is_valid(self, action_data: ActionDataT) -> bool:
        value = getattr(action_data.get(self.property_object), self.attribute)
        if self.equal_to is self.SOMETHING:
            return bool(value)
        return value is self.equal_to


class ConditionSet(Condition, SystemActionTag):
    """
    Set of conditions
    """

    CONDITIONS: Optional[Iterable[Condition]] = None

    def __init__(
        self,
        conditions: Optional[Iterable[Condition]] = None,
        fail_message: Optional[str] = None,
        name: Optional[str] = None,
    ):
        self.conditions = conditions or self.CONDITIONS or raise_not_provided_field("conditions")

        super().__init__(
            fail_message=fail_message or "({})".format(", ".join([cond.fail_message for cond in self.conditions])),
            name="{}: {}".format(name or self.NAME, "AND ".join(cond.action_name for cond in self.conditions)),
        )

    @property
    def action_name(self) -> str:
        return "{}[{}]".format(
            self.name_with_inverted, "AND ".join(cond.name_with_inverted for cond in self.conditions)
        )

    def _is_valid(self, action_data: ActionDataT) -> bool:
        return all(cond.validate(action_data) for cond in self.conditions)


class If(ConditionSet, SystemActionTag):
    """
    Syntactic sugar for `ConditionSet`. See `Switch` docs for usage.
    """

    def __init__(self, *conditions: Condition):
        super().__init__(conditions)


class Switch(Action, SystemActionTag):
    """
    Branching action which can be used in three ways:

    1) Create at once via constructor by pairs of condition-action
    2) By `if_then` chaining: Switch().if_then(cond1, action1).if_then(cond2, action2).otherwise(action3)
    3) Use `Switch.case` with `If` and `Then`:

    .. code-block:: python

        Switch()
        .case(
            If(cond.ClaimInValidState("PAID")),
            Then(
                ~cond.ClaimHasPaymentAuthorizationAssigned(
                    fail_message="You cannot reset a claim that has payment authorizations assigned!"
                ),
                cond.ClaimFromPaymentPlatform(),
                trans.StateToAuthorize(),
                trans.ResetEligibleAmount(),
            ),
        )
        .case(
            If(
                cond.ClaimInValidState("DECLINED", "CANCELLED"),
                cond.ClaimHasPreAuth(),
                ~cond.ClaimHasValidNotification(),
                ~cond.ClaimIsRetroactive(),
            ),
            Then(cond.UserCanResetClaim(), trans.StateToPendingAuthorization()),
        )
    """

    def __init__(
        self,
        paths: Optional[List[Tuple[Condition, Optional[ActionT]]]] = None,
        fail_message: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        :param paths: Pairs of condition and action which is executed whether condition is met
        :param fail_message:
        """
        super().__init__(description=fail_message, name=name)
        self.paths = paths or []

    @record_action
    @verbose_action_exception
    def run(self, action_data: ActionDataT) -> ActionDataT:
        for condition, action in self.paths:
            if action_data.skip_processing:
                return action_data

            action_data = action_data.record_start(condition)
            validation = condition.validate(action_data)
            action_data = action_data.record_end(condition)

            if validation:
                if not action:
                    return action_data
                return action.run(action_data)

        raise NoneOfActionsCanBeExecuted(
            "{}. {}".format(
                self.description, json.dumps([(p[0].action_name, p[0].description) for p in self.paths], indent=4)
            )
        )

    @async_record_action
    @async_verbose_action_exception
    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        for condition, action in self.paths:
            if action_data.skip_processing:
                return action_data

            action_data = action_data.record_start(condition)
            validation = condition.validate(action_data)
            action_data = action_data.record_end(condition)

            if validation:
                if not action:
                    return action_data
                return await action.async_run(action_data)

        raise NoneOfActionsCanBeExecuted(
            "{}. {}".format(
                self.description, json.dumps([(p[0].action_name, p[0].description) for p in self.paths], indent=4)
            )
        )

    def if_then(self, condition: Condition, action: Optional[ActionT] = None) -> "Switch":
        """
        Add a condition-action pair

        :param condition:
        :param action:
        :return: New `Switch` instance with appended condition-action pair
        """
        return Switch(self.paths + [(condition, action)], fail_message=self.description)

    def case(self, if_condition: If, then_action: Optional[Then] = None) -> "Switch":
        return self.if_then(if_condition, then_action)

    def otherwise(self, action: Optional[ActionT] = None) -> "Switch":
        return self.if_then(AlwaysTrue(), action)


class NonNoneDataValues(Condition):
    """
    Condition which determines whether certain values are not `None`
    """

    def __init__(self, *fields: str):
        """
        :param fields: Fields of :class:`~orinoco.entities.ActionData` which will be checked if are present and not none
        """
        super().__init__(
            fail_message="Data has to contain following non-none fields: {}".format(fields),
            name="{}[{}]".format(self.__class__.__name__, ", ".join(fields)),
        )
        self.fields = fields

    def _is_valid(self, action_data: ActionDataT) -> bool:
        return all(action_data.is_in(field) and action_data.get(field) is not None for field in self.fields)


class BaseOperator(ConditionSet, ABC):
    """
    Base operator between two conditions
    """

    def __init__(self, cond1: Condition, cond2: Condition, description: Optional[str] = None):
        super().__init__(
            [cond1, cond2],
            fail_message=description or "Operator failed for {} and {}".format(cond1.action_name, cond2.action_name),
        )


class AndOperator(BaseOperator, SystemActionTag):
    def _is_valid(self, action_data: ActionDataT) -> bool:
        return all(cond.validate(action_data) for cond in self.conditions)


class OrOperator(BaseOperator, SystemActionTag):

    FAIL_MESSAGE = "None of conditions is True"

    def _is_valid(self, action_data: ActionDataT) -> bool:
        return any(cond.validate(action_data) for cond in self.conditions)


class ConditionalAction(Action, SystemActionTag):
    """
    Represent action which is executed only when the given condition is met
    """

    def __init__(self, condition: Condition, action: ActionT, name: Optional[str] = None):
        super().__init__(name=name or "If({}) -> {}".format(condition.action_name, action.action_name))
        self.condition = condition
        self.action = action

    def run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        if self.condition.validate(action_data):
            return self.action.run(action_data)
        return action_data

    async def async_run(self, action_data: ActionDataT) -> ActionDataT:
        if action_data.skip_processing:
            return action_data

        if self.condition.validate(action_data):
            return await self.action.async_run(action_data)
        return action_data


class IsInData(Condition):
    """
    Condition which checks whether a value with the given key exists in the data container
    """

    def __init__(self, field: str):
        super().__init__(
            fail_message="Field {} is not in data".format(field), name="{}[{}]".format(self.__class__.__name__, field)
        )
        self.field = field

    def _is_valid(self, action_data: ActionDataT) -> bool:
        return action_data.is_in(self.field)


class SignatureIsInData(Generic[T], Condition):
    """
    Condition which checks whether a value with the given signature exists in the data container
    """

    def __init__(self, signature: Signature[T]):
        super().__init__(
            fail_message="Field {} is not in data".format(str(signature)),
            name="{}[{}]".format(self.__class__.__name__, signature),
        )
        self.signature = signature

    def _is_valid(self, action_data: ActionDataT) -> bool:
        return action_data.signature_is_in(self.signature)


class LoopCondition(Condition, SystemActionTag):
    """
    Condition which checks whether values in an iterable item match the given condition
    """

    def __init__(
        self,
        method: Callable[[Union[List[bool], Generator[bool, None, None]]], bool],
        iterable_key: str,
        as_key: str,
        condition: Condition,
        fail_message: Optional[str] = None,
        name: Optional[str] = None,
    ):
        self.method = method
        self.iterable_key = iterable_key
        self.as_key = as_key
        self.condition = condition
        super().__init__(fail_message=fail_message, name=name)

    def _is_valid(self, action_data: ActionDataT) -> bool:
        return self.method(self._get_generator(action_data))

    def _get_generator(self, action_data: ActionDataT) -> Generator[bool, None, None]:
        return (
            self.condition.validate(action_data.evolve(**{self.as_key: value}))
            for value in action_data.get(self.iterable_key)
        )


class AnyCondition(LoopCondition, SystemActionTag):
    """
    Condition which checks whether any value in an iterable item match the given condition
    """

    def __init__(self, iterable_key: str, as_key: str, condition: Condition):
        super().__init__(
            method=any,
            iterable_key=iterable_key,
            as_key=as_key,
            condition=condition,
            fail_message="Any of the given items meet the condition [{}: {}]".format(
                condition.__class__.__name__, condition.fail_message
            ),
        )


class AllCondition(LoopCondition, SystemActionTag):
    """
    Condition which checks whether all values in an iterable item match the given condition
    """

    def __init__(self, iterable_key: str, as_key: str, condition: Condition):
        super().__init__(
            method=all,
            iterable_key=iterable_key,
            as_key=as_key,
            condition=condition,
            fail_message="All of the given items meet the condition [{}: {}]".format(
                condition.__class__.__name__, condition.fail_message
            ),
        )
